@@ -16,22 +16,33 @@
 from __future__ import division, absolute_import
 
 import functools
+import platform
 import six
 import struct
 import weakref
 
-# We're just re-exporting visualtype_to_c_struct, hence the noqa.
-from .ffi import ffi, C, visualtype_to_c_struct  # noqa
+try:
+    from xcffib._ffi import ffi
+except ImportError:
+    from xcffib.ffi_build import ffi
 
-__xcb_proto_version__ = "1.10"
+soname = "libxcb.so.1"
+if platform.system() == "Darwin":
+    soname = "libxcb.dylib"
+elif platform.system() == "Windows":
+    soname = "libxcb.dll"
+lib = ffi.dlopen(soname)
 
-X_PROTOCOL = C.X_PROTOCOL
-X_PROTOCOL_REVISION = C.X_PROTOCOL_REVISION
+__xcb_proto_version__ = "1.11"
+__version__ = "0.5.1"
 
-XCB_NONE = C.XCB_NONE
-XCB_COPY_FROM_PARENT = C.XCB_COPY_FROM_PARENT
-XCB_CURRENT_TIME = C.XCB_CURRENT_TIME
-XCB_NO_SYMBOL = C.XCB_NO_SYMBOL
+X_PROTOCOL = lib.X_PROTOCOL
+X_PROTOCOL_REVISION = lib.X_PROTOCOL_REVISION
+
+XCB_NONE = lib.XCB_NONE
+XCB_COPY_FROM_PARENT = lib.XCB_COPY_FROM_PARENT
+XCB_CURRENT_TIME = lib.XCB_CURRENT_TIME
+XCB_NO_SYMBOL = lib.XCB_NO_SYMBOL
 
 # For xpyb compatibility
 NONE = XCB_NONE
@@ -39,13 +50,13 @@ CopyFromParent = XCB_COPY_FROM_PARENT
 CurrentTime = XCB_CURRENT_TIME
 NoSymbol = XCB_NO_SYMBOL
 
-XCB_CONN_ERROR = C.XCB_CONN_ERROR
-XCB_CONN_CLOSED_EXT_NOTSUPPORTED = C.XCB_CONN_CLOSED_EXT_NOTSUPPORTED
-XCB_CONN_CLOSED_MEM_INSUFFICIENT = C.XCB_CONN_CLOSED_MEM_INSUFFICIENT
-XCB_CONN_CLOSED_REQ_LEN_EXCEED = C.XCB_CONN_CLOSED_REQ_LEN_EXCEED
-XCB_CONN_CLOSED_PARSE_ERR = C.XCB_CONN_CLOSED_PARSE_ERR
-# XCB_CONN_CLOSED_INVALID_SCREEN = C.XCB_CONN_CLOSED_INVALID_SCREEN
-# XCB_CONN_CLOSED_FDPASSING_FAILED = C.XCB_CONN_CLOSED_FDPASSING_FAILED
+XCB_CONN_ERROR = lib.XCB_CONN_ERROR
+XCB_CONN_CLOSED_EXT_NOTSUPPORTED = lib.XCB_CONN_CLOSED_EXT_NOTSUPPORTED
+XCB_CONN_CLOSED_MEM_INSUFFICIENT = lib.XCB_CONN_CLOSED_MEM_INSUFFICIENT
+XCB_CONN_CLOSED_REQ_LEN_EXCEED = lib.XCB_CONN_CLOSED_REQ_LEN_EXCEED
+XCB_CONN_CLOSED_PARSE_ERR = lib.XCB_CONN_CLOSED_PARSE_ERR
+# XCB_CONN_CLOSED_INVALID_SCREEN = lib.XCB_CONN_CLOSED_INVALID_SCREEN
+# XCB_CONN_CLOSED_FDPASSING_FAILED = lib.XCB_CONN_CLOSED_FDPASSING_FAILED
 
 cffi_explicit_lifetimes = weakref.WeakKeyDictionary()
 
@@ -54,6 +65,22 @@ def type_pad(t, i):
     return -i & (3 if t > 4 else t - 1)
 
 
+def visualtype_to_c_struct(vt):
+    # let ffi be a kwarg so cairocffi can pass in its ffi
+    # cfficairo needs an xcb_visualtype_t
+    s = ffi.new("struct xcb_visualtype_t *")
+
+    s.visual_id = vt.visual_id
+    s._class = vt._class
+    s.bits_per_rgb_value = vt.bits_per_rgb_value
+    s.colormap_entries = vt.colormap_entries
+    s.red_mask = vt.red_mask
+    s.green_mask = vt.green_mask
+    s.blue_mask = vt.blue_mask
+
+    return s
+
+
 class Unpacker(object):
 
     def __init__(self, known_max=None):
@@ -64,8 +91,7 @@ class Unpacker(object):
             self._resize(known_max)
 
     def pad(self, thing):
-        if isinstance(thing, type) and any(
-                [issubclass(thing, c) for c in [Struct, Union]]):
+        if isinstance(thing, type) and issubclass(thing, (Struct, Union)):
             if hasattr(thing, "fixed_size"):
                 size = thing.fixed_size
             else:
@@ -123,9 +149,9 @@ class CffiUnpacker(Unpacker):
 
 class MemoryUnpacker(Unpacker):
 
-    def __init__(self, data, fmt):
-        self.buf = struct.pack(fmt, *data)
-        Unpacker.__init__(self, struct.calcsize(fmt))
+    def __init__(self, buf):
+        self.buf = buf
+        Unpacker.__init__(self, len(self.buf))
 
     def _resize(self, increment):
         if self.size + increment > self.known_max:
@@ -133,11 +159,9 @@ class MemoryUnpacker(Unpacker):
         self.size += increment
 
     def copy(self):
-        new = MemoryUnpacker.__new__(MemoryUnpacker)
-        new.buf = self.buf
+        new = MemoryUnpacker(self.buf)
         new.offset = self.offset
         new.size = self.size
-        new.known_max = self.known_max
         return new
 
 
@@ -153,23 +177,23 @@ class XcffibException(Exception):
 
 class ConnectionException(XcffibException):
     REASONS = {
-        C.XCB_CONN_ERROR: (
+        lib.XCB_CONN_ERROR: (
             'xcb connection errors because of socket, '
             'pipe and other stream errors.'),
-        C.XCB_CONN_CLOSED_EXT_NOTSUPPORTED: (
+        lib.XCB_CONN_CLOSED_EXT_NOTSUPPORTED: (
             'xcb connection shutdown because extension not supported'),
-        C.XCB_CONN_CLOSED_MEM_INSUFFICIENT: (
+        lib.XCB_CONN_CLOSED_MEM_INSUFFICIENT: (
             'malloc(), calloc() and realloc() error upon failure, '
             'for eg ENOMEM'),
-        C.XCB_CONN_CLOSED_REQ_LEN_EXCEED: (
+        lib.XCB_CONN_CLOSED_REQ_LEN_EXCEED: (
             'Connection closed, exceeding request length that server '
             'accepts.'),
-        C.XCB_CONN_CLOSED_PARSE_ERR: (
+        lib.XCB_CONN_CLOSED_PARSE_ERR: (
             'Connection closed, error during parsing display string.'),
-        #        C.XCB_CONN_CLOSED_INVALID_SCREEN: (
+        #        lib.XCB_CONN_CLOSED_INVALID_SCREEN: (
         #            'Connection closed because the server does not have a screen '
         #            'matching the display.'),
-        #        C.XCB_CONN_CLOSED_FDPASSING_FAILED: (
+        #        lib.XCB_CONN_CLOSED_FDPASSING_FAILED: (
         #            'Connection closed because some FD passing operation failed'),
     }
 
@@ -185,7 +209,10 @@ class ProtocolException(XcffibException):
 core = None
 core_events = None
 core_errors = None
-setup = None
+# we use _setup here instead of just setup because of a nose bug that triggers
+# when doing the packaging builds in debian:
+# https://code.google.com/p/python-nose/issues/detail?id=326
+_setup = None
 
 extensions = {}
 
@@ -193,22 +220,22 @@ extensions = {}
 # a core besides xproto, so why not just hardcode that?
 
 
-def _add_core(value, _setup, events, errors):
+def _add_core(value, __setup, events, errors):
     if not issubclass(value, Extension):
         raise XcffibException(
             "Extension type not derived from xcffib.Extension")
-    if not issubclass(_setup, Struct):
+    if not issubclass(__setup, Struct):
         raise XcffibException("Setup type not derived from xcffib.Struct")
 
     global core
     global core_events
     global core_errors
-    global setup
+    global _setup
 
     core = value
     core_events = events
     core_errors = errors
-    setup = _setup
+    _setup = __setup
 
 
 def _add_ext(key, value, events, errors):
@@ -251,9 +278,11 @@ class Protobj(object):
     """ Note: Unlike xcb.Protobj, this does NOT implement the sequence
     protocol. I found this behavior confusing: Protobj would implement the
     sequence protocol on self.buf, and then List would go and implement it on
-    List. Additionally, as near as I can tell internally we only need the size
-    of the buffer for cases when the size of things is unspecified. Thus,
-    that's all we save.
+    List.
+
+    Instead, when we need to create a new event from an existing event, we
+    repack that event into a MemoryUnpacker and use that instead (see
+    eventToUnpacker in the generator for more info.)
     """
 
     def __init__(self, unpacker):
@@ -267,13 +296,6 @@ class Protobj(object):
         if unpacker.known_max is not None:
             self.bufsize = unpacker.known_max
 
-    @classmethod
-    def synthetic(cls, **kwargs):
-        self = cls.__new__(cls)
-        for k, v in kwargs.items():
-            setattr(self, k, v)
-        return self
-
 
 class Struct(Protobj):
     pass
@@ -283,7 +305,7 @@ class Union(Protobj):
     @classmethod
     def synthetic(cls, data=[], fmt=""):
         self = cls.__new__(cls)
-        self.__init__(MemoryUnpacker(data, fmt))
+        self.__init__(MemoryUnpacker(struct.pack(fmt, *data)))
         return self
 
 
@@ -305,6 +327,9 @@ class Cookie(object):
             "Request is not void and checked")
         self.conn.request_check(self.sequence)
 
+    def discard_reply(self):
+        return self.conn.discard_reply(self.sequence)
+
 
 class VoidCookie(Cookie):
 
@@ -350,7 +375,7 @@ class Extension(object):
         xcb_parts[3].iov_base = ffi.NULL
         xcb_parts[3].iov_len = -len(data) & 3  # is this really necessary?
 
-        flags = C.XCB_REQUEST_CHECKED if is_checked else 0
+        flags = lib.XCB_REQUEST_CHECKED if is_checked else 0
 
         seq = self.conn.send_request(flags, xcb_parts + 2, xcb_req)
 
@@ -424,11 +449,20 @@ class List(Protobj):
 
     def to_atoms(self):
         """ A helper for converting a List of chars to an array of atoms """
-        return struct.unpack("=" + "I" * (len(self) // 4), b''.join(self))
+        return struct.unpack("<" + "I" * (len(self) // 4), b''.join(self))
 
     def buf(self):
         return b''.join(self.list)
 
+    @classmethod
+    def synthetic(cls, list=None):
+        if list is None:
+            list = []
+
+        self = cls.__new__(cls)
+        self.list = list[:]
+        return self
+
 
 class OffsetMap(object):
 
@@ -471,13 +505,19 @@ class Connection(object):
         i = ffi.new("int *")
 
         if fd > 0:
-            self._conn = C.xcb_connect_to_fd(fd, c_auth)
+            self._conn = lib.xcb_connect_to_fd(fd, c_auth)
         elif c_auth != ffi.NULL:
-            self._conn = C.xcb_connect_to_display_with_auth_info(display, c_auth, i)
+            self._conn = lib.xcb_connect_to_display_with_auth_info(display, c_auth, i)
         else:
-            self._conn = C.xcb_connect(display, i)
+            self._conn = lib.xcb_connect(display, i)
         self.pref_screen = i[0]
         self.invalid()
+        self._init_x()
+
+    def _init_x(self):
+        if core is None:
+            raise XcffibException("No core protocol object has been set.  "
+                                  "Did you import xcffib.xproto?")
 
         self.core = core(self)
         self.setup = self.get_setup()
@@ -491,7 +531,7 @@ class Connection(object):
             # We're explicitly not putting this as an argument to the next call
             # as a hack for lifetime management.
             c_ext = key.to_cffi()
-            reply = C.xcb_get_extension_data(self._conn, c_ext)
+            reply = lib.xcb_get_extension_data(self._conn, c_ext)
             self._event_offsets.add(reply.first_event, events)
             self._error_offsets.add(reply.first_error, errors)
 
@@ -501,7 +541,7 @@ class Connection(object):
     def invalid(self):
         if self._conn is None:
             raise XcffibException("Invalid connection.")
-        err = C.xcb_connection_has_error(self._conn)
+        err = lib.xcb_connection_has_error(self._conn)
         if err > 0:
             raise ConnectionException(err)
 
@@ -522,57 +562,70 @@ class Connection(object):
 
     @ensure_connected
     def get_setup(self):
-        s = C.xcb_get_setup(self._conn)
+        self._setup = lib.xcb_get_setup(self._conn)
 
         # No idea where this 8 comes from either, similar complate to the
         # sizeof(xcb_generic_reply_t) below.
-        buf = CffiUnpacker(s, known_max=8 + s.length * 4)
+        buf = CffiUnpacker(self._setup, known_max=8 + self._setup.length * 4)
 
-        return setup(buf)
+        return _setup(buf)
+
+    @ensure_connected
+    def get_screen_pointers(self):
+        """
+        Returns the xcb_screen_t for every screen
+        useful for other bindings
+        """
+        root_iter = lib.xcb_setup_roots_iterator(self._setup)
+
+        screens = [root_iter.data]
+        for i in range(self._setup.roots_len - 1):
+            lib.xcb_screen_next(ffi.addressof((root_iter)))
+            screens.append(root_iter.data)
+        return screens
 
     @ensure_connected
     def wait_for_event(self):
-        e = C.xcb_wait_for_event(self._conn)
-        e = ffi.gc(e, C.free)
+        e = lib.xcb_wait_for_event(self._conn)
+        e = ffi.gc(e, lib.free)
         self.invalid()
         return self.hoist_event(e)
 
     @ensure_connected
     def poll_for_event(self):
-        e = C.xcb_poll_for_event(self._conn)
+        e = lib.xcb_poll_for_event(self._conn)
         self.invalid()
         if e != ffi.NULL:
             return self.hoist_event(e)
         else:
             return None
 
-    @ensure_connected
     def has_error(self):
-        return C.xcb_connection_has_error(self._conn)
+        return lib.xcb_connection_has_error(self._conn)
 
     @ensure_connected
     def get_file_descriptor(self):
-        return C.xcb_get_file_descriptor(self._conn)
+        return lib.xcb_get_file_descriptor(self._conn)
 
     @ensure_connected
     def get_maximum_request_length(self):
-        return C.xcb_get_maximum_request_length(self._conn)
+        return lib.xcb_get_maximum_request_length(self._conn)
 
     @ensure_connected
     def prefetch_maximum_request_length(self):
-        return C.xcb_prefetch_maximum_request_length(self._conn)
+        return lib.xcb_prefetch_maximum_request_length(self._conn)
 
     @ensure_connected
     def flush(self):
-        return C.xcb_flush(self._conn)
+        return lib.xcb_flush(self._conn)
 
     @ensure_connected
     def generate_id(self):
-        return C.xcb_generate_id(self._conn)
+        return lib.xcb_generate_id(self._conn)
 
     def disconnect(self):
         self.invalid()
-        return C.xcb_disconnect(self._conn)
+        return lib.xcb_disconnect(self._conn)
 
     def _process_error(self, c_error):
         self.invalid()
@@ -584,14 +637,14 @@ class Connection(object):
     @ensure_connected
     def wait_for_reply(self, sequence):
         error_p = ffi.new("xcb_generic_error_t **")
-        data = C.xcb_wait_for_reply(self._conn, sequence, error_p)
-        data = ffi.gc(data, C.free)
+        data = lib.xcb_wait_for_reply(self._conn, sequence, error_p)
+        data = ffi.gc(data, lib.free)
 
         try:
             self._process_error(error_p[0])
         finally:
             if error_p[0] != ffi.NULL:
-                C.free(error_p[0])
+                lib.free(error_p[0])
 
         if data == ffi.NULL:
             # No data and no error => bad sequence number
@@ -599,7 +652,10 @@ class Connection(object):
 
         reply = ffi.cast("xcb_generic_reply_t *", data)
 
-        # why is this 32 and not sizeof(xcb_generic_reply_t) == 8?
+        # this is 32 and not `sizeof(xcb_generic_reply_t) == 8` because,
+        # according to the X11 protocol specs: "Every reply consists of 32 bytes
+        # followed by zero or more additional bytes of data, as specified in the
+        # length field."
         return CffiUnpacker(data, known_max=32 + reply.length * 4)
 
     @ensure_connected
@@ -607,7 +663,7 @@ class Connection(object):
         cookie = ffi.new("xcb_void_cookie_t [1]")
         cookie[0].sequence = sequence
 
-        err = C.xcb_request_check(self._conn, cookie[0])
+        err = lib.xcb_request_check(self._conn, cookie[0])
         self._process_error(err)
 
     def hoist_event(self, e):
@@ -626,7 +682,11 @@ class Connection(object):
 
     @ensure_connected
     def send_request(self, flags, xcb_parts, xcb_req):
-        return C.xcb_send_request(self._conn, flags, xcb_parts, xcb_req)
+        return lib.xcb_send_request(self._conn, flags, xcb_parts, xcb_req)
+
+    @ensure_connected
+    def discard_reply(self, sequence):
+        return lib.xcb_discard_reply(self._conn, sequence)
 
 
 # More backwards compatibility
@@ -641,9 +701,17 @@ class Response(Protobj):
         # These (and the ones in Reply) aren't used internally and I suspect
         # they're not used by anyone else, but they're here for xpyb
         # compatibility.
-        resp = unpacker.cast("xcb_generic_event_t *")
-        self.response_type = resp.response_type
-        self.sequence = resp.sequence
+        #
+        # In some cases (e.g. creating synthetic events from memory), we don't
+        # have the sequence number (since the event was fake), so only try to
+        # get these attributes if we are really using a cffi buffer.
+        if isinstance(unpacker, CffiUnpacker):
+            resp = unpacker.cast("xcb_generic_event_t *")
+            self.response_type = resp.response_type
+            self.sequence = resp.sequence
+        else:
+            self.response_type = None
+            self.sequence = None
 
 
 class Reply(Response):
@@ -703,9 +771,25 @@ def pack_list(from_, pack_type):
     else:
         buf = six.BytesIO()
         for item in from_:
-            # If we can't pack it, you'd better have packed it yourself...
-            if isinstance(item, Struct):
+            # If we can't pack it, you'd better have packed it yourself. But
+            # let's not confuse things which aren't our Probobjs for packable
+            # things.
+            if isinstance(item, Protobj) and hasattr(item, "pack"):
                 buf.write(item.pack())
             else:
                 buf.write(item)
         return buf.getvalue()
+
+
+def wrap(ptr):
+    c_conn = ffi.cast('xcb_connection_t *', ptr)
+    conn = Connection.__new__(Connection)
+    conn._conn = c_conn
+    conn._init_x()
+    conn.invalid()
+
+    # ptr owns the memory for c_conn, even after the cast
+    # we should keep it alive
+    cffi_explicit_lifetimes[conn] = ptr
+
+    return conn
